from .linebotx import LineBotApiAsync, WebhookHandlerAsync, HttpXClient
from quart import Quart, request, abort, Response, send_from_directory
from linebot.exceptions import InvalidSignatureError
from linebot.models import MessageEvent, TextMessage, TextSendMessage, ImageMessage, ImageSendMessage, StickerMessage, \
    StickerSendMessage, VideoMessage, AudioMessage
import asyncio
from typing import Dict, DefaultDict, Union
from typing_extensions import Literal
from pydantic import Field
from io import BytesIO
from PIL import Image
from unified_message_relay.Core import UMRLogging
from unified_message_relay.Core import UMRDriver
from unified_message_relay.Core import UMRConfig
from unified_message_relay.Core.UMRType import UnifiedMessage, MessageEntity, ChatType, EntityType
from unified_message_relay.Core.UMRMessageRelation import set_ingress_message_id, set_egress_message_id
from unified_message_relay.Util.Helper import unparse_entities_to_markdown
import threading
import os
from uuid import uuid4
from collections import defaultdict
from PIL import ImageFile


class LineDriverConfig(UMRConfig.BaseDriverConfig):
    Base: Literal['Line']
    ChannelID: str
    BotToken: str
    WebHookToken: str
    WebHookURL: str
    WebHookPort: int = Field(28080, ge=0, le=65535)
    HTTPSCert: str
    HTTPSKey: str
    HTTPSCA: str


UMRConfig.register_driver_config(LineDriverConfig)


class LineDriver(UMRDriver.BaseDriverMixin):
    def __init__(self, name):
        super().__init__(name)

        self.name = name
        self.logger = UMRLogging.get_logger(f'UMRDriver.{self.name}')
        self.logger.debug(f'Started initialization for {self.name}')
        self.loop: asyncio.AbstractEventLoop = asyncio.new_event_loop()
        self.loop.set_exception_handler(self.handle_exception)
        self.config: LineDriverConfig = UMRConfig.config.Driver[self.name]
        self.handler: WebhookHandlerAsync

        self.user_names: Dict[str, str] = dict()  # [user_id, username]
        self._message_id = 0
        self.data_root = UMRConfig.config.DataRoot
        self.channel_id = self.config.ChannelID
        self.image_webhook_url = self.config.WebHookURL + ':' + str(self.config.WebHookPort) + '/image/'

        # mapping between internal type and external type
        self.chat_type_dict = {
            'user': ChatType.PRIVATE,
            'room': ChatType.DISCUSS,
            'group': ChatType.GROUP
        }

        self.chat_type_dict_reversed = {v: k for k, v in self.chat_type_dict.items()}

        self.loop = asyncio.new_event_loop()

        transport = HttpXClient()
        self.bot = LineBotApiAsync(self.config.BotToken, http_client=transport)
        self.app = Quart(__name__)
        self.handler = WebhookHandlerAsync(self.config.WebHookToken)

        # previous reply token, try to save push api limit
        self.reply_token: DefaultDict[str, str] = defaultdict(lambda: '')

        @self.app.route("/callback", methods=['POST'])
        async def callback():
            # get X-Line-Signature header value
            signature = request.headers['X-Line-Signature']

            data = await request.get_data()

            body = data.decode()
            print("Request body: " + body)

            # handle webhook body
            try:
                await self.handler.handle(body, signature)
            except InvalidSignatureError:
                abort(400)

            return Response('OK')

        @self.app.route('/image/<path:path>')
        async def image_hosting(path):
            return await send_from_directory(self.data_root, path)

        @self.handler.add(MessageEvent)
        async def message_text(event: MessageEvent):
            if event.source.type == 'user':
                chat_id = event.source.user_id
                username = await self.get_user_name(user_id=event.source.user_id)
            elif event.source.type == 'group':
                chat_id = event.source.group_id
                username = await self.get_user_name(user_id=event.source.user_id, group_id=event.source.group_id)
            elif event.source.type == 'room':
                chat_id = event.source.room_id
                username = await self.get_user_name(user_id=event.source.user_id, room_id=event.source.room_id)
            else:  # unknown source
                return

            self.reply_token[chat_id] = event.reply_token

            _chat_type = self.chat_type_dict[event.source.type]
            pseudo_message_id = self.message_id
            message = UnifiedMessage(platform=self.name, chat_id=chat_id, chat_type=_chat_type, name=username, message_id=pseudo_message_id)
            set_ingress_message_id(src_platform=self.name, src_chat_id=chat_id, src_chat_type=_chat_type,
                                   src_message_id=pseudo_message_id, user_id=0)
            if isinstance(event.message, TextMessage):
                message.text = event.message.text
            elif isinstance(event.message, StickerMessage):
                message.text = 'Sent a sticker'
            elif isinstance(event.message, ImageMessage):
                message_content = await self.bot.get_message_content(event.message.id)
                image_content = message_content.content
                image: Image.Image = Image.open(BytesIO(image_content))
                image_path = os.path.join(self.data_root, str(uuid4()) + '.jpg')
                image.save(image_path)
                message.image = image_path
            else:
                message.text = 'Unsupported message type'
            await self.receive(message)

    @property
    def message_id(self):
        self._message_id += 1
        return self._message_id

    async def get_user_name(self, user_id, group_id=None, room_id=None):
        """
        get user name by get_profile
        :param user_id: raw user id
        :param group_id: if not None, call get_group_member_profile
        :param room_id: if not None, call get_room_member_profile
        :return: str
        """
        if user_id in self.user_names:
            return self.user_names[user_id]

        if group_id is not None:
            profile = await self.bot.get_group_member_profile(group_id=group_id, user_id=user_id)
        elif room_id is not None:
            profile = await self.bot.get_room_member_profile(room_id=room_id, user_id=user_id)
        else:
            profile = await self.bot.get_profile(user_id=user_id)

        self.user_names[user_id] = profile.display_name
        return profile.display_name

    def start(self):

        def run():
            nonlocal self
            asyncio.set_event_loop(self.loop)

            task = self.app.run_task(host='0.0.0.0', port=self.config.WebHookPort,
                                     ca_certs=self.config.HTTPSCA,
                                     keyfile=self.config.HTTPSKey,
                                     certfile=self.config.HTTPSCert)

            self.loop.create_task(task)
            self.loop.run_forever()

        t = threading.Thread(target=run)
        t.daemon = True
        UMRDriver.threads.append(t)
        t.start()

        self.logger.debug(f'Finished initialization for {self.name}')

    async def send(self, to_chat: Union[int, str], chat_type: ChatType, messsage: UnifiedMessage):
        """
        decorator for send new message
        :return:
        """
        self.logger.debug('calling real send')
        return asyncio.run_coroutine_threadsafe(self._send(to_chat, chat_type, messsage), self.loop)

    async def _send(self, to_chat: str, chat_type: ChatType, message: UnifiedMessage):
        """
        decorator for send new message
        :return:
        """
        self.logger.debug('begin processing message')

        _chat_type = self.chat_type_dict_reversed[chat_type]

        message_prefix = ''
        if message.chat_attrs.name:
            message_prefix = message.chat_attrs.name
        if message.chat_attrs.reply_to:
            message_prefix += ' (➡️️' + message.chat_attrs.reply_to.name + ')'
        if message.chat_attrs.forward_from:
            message_prefix += ' (️️↩️' + message.chat_attrs.forward_from.name + ')'
        if message.chat_attrs.name:
            message_prefix += ': '

        if message.image:
            _, original_file_name = os.path.split(message.image)
            file_name, file_ext = original_file_name.split('.')
            image_original = file_name + '-origin.' + file_ext
            image_thumb = file_name + '-thumb.' + file_ext
            image_original_path = os.path.join(self.data_root, image_original)
            image_thumb_path = os.path.join(self.data_root, image_thumb)

            if not os.path.isfile(image_original_path):
                image: Image.Image = Image.open(message.image)
                if image.size[0] <= 1024 and image.size[1] <= 1024:
                    image_original = original_file_name
                else:
                    image.thumbnail((1024, 1024), Image.ANTIALIAS)
                    image.save(image_original_path)
                image.thumbnail((240, 240), Image.ANTIALIAS)
                image.save(image_thumb_path)

            if self.reply_token[to_chat]:
                await self.bot.reply_message(reply_token=self.reply_token[to_chat],
                                             messages=TextSendMessage(text=message_prefix + 'Sent an image ⬇️'))
                self.logger.debug('Begin sending image')
                await self.bot.reply_message(reply_token=self.reply_token[to_chat],
                                             messages=ImageSendMessage(
                                                 original_content_url=self.image_webhook_url + image_original,
                                                 preview_image_url=self.image_webhook_url + image_thumb)
                                             )
                self.logger.debug('Finished sending image')
            else:
                await self.bot.push_message(to=to_chat, messages=TextSendMessage(text=message_prefix + 'Sent an image ⬇️'))
                self.logger.debug('Begin sending image')
                await self.bot.push_message(to=to_chat,
                                            messages=ImageSendMessage(original_content_url=self.image_webhook_url + image_original,
                                                                      preview_image_url=self.image_webhook_url + image_thumb)
                )
                self.logger.debug('Finished sending image')
        if message.text:
            message_text = unparse_entities_to_markdown(message, EntityType.PLAIN)
            _message = TextSendMessage(text=message_prefix + message_text)
            if self.reply_token[to_chat]:
                await self.bot.reply_message(reply_token=self.reply_token[to_chat], messages=_message)
            else:
                await self.bot.push_message(to=to_chat, messages=_message)
        if message.chat_attrs:
            set_egress_message_id(src_platform=message.chat_attrs.platform,
                                  src_chat_id=message.chat_attrs.chat_id,
                                  src_chat_type=message.chat_attrs.chat_type,
                                  src_message_id=message.chat_attrs.message_id,
                                  dst_platform=self.name,
                                  dst_chat_id=to_chat,
                                  dst_chat_type=_chat_type,
                                  dst_message_id=self.message_id,  # useless message id
                                  user_id=0)

    def handle_exception(self, loop, context):
        # e = context.get('exception')
        # if isinstance(e, SSLError):
        #     return
        msg = context.get("exception", context["message"])
        self.logger.exception('Unhandled exception: ', exc_info=msg)

    async def is_group_admin(self, chat_id: int, chat_type: ChatType, user_id: int):
        return False

    async def is_group_owner(self, chat_id: int, chat_type: ChatType, user_id: int):
        return False


UMRDriver.register_driver('Line', LineDriver)



import asyncio

import httpx

from linebot.http_client import HttpClient, HttpResponse
import json


class HttpXClient(HttpClient):
    """HttpClient implemented by httpx."""

    DEFAULT_TIMEOUT = 30

    def __init__(self, timeout=DEFAULT_TIMEOUT):
        self.session = httpx.AsyncClient(timeout=timeout)
        self.timeout = timeout

    async def get(self, url, headers=None, params=None, stream=False, timeout=None):
        if timeout is None:
            timeout = self.timeout

        response = await self.session.get(url, headers=headers, params=params, timeout=timeout)
        return AioHttpResponse(response)

    async def post(self, url, headers=None, data=None, timeout=None):
        if timeout is None:
            timeout = self.timeout

        response = await self.session.post(url, headers=headers, data=data, timeout=timeout)
        return AioHttpResponse(response)

    async def delete(self, url, headers=None, params=None, timeout=None):
        if timeout is None:
            timeout = self.timeout

        response = await self.session.delete(
            url, headers=headers, params=params, timeout=timeout
        )

        return AioHttpResponse(response)

    async def close(self):
        await self.session.aclose()


class AioHttpResponse(HttpResponse):
    """HttpResponse implemented by aiohttp lib's response."""

    def __init__(self, response: httpx.Response):
        self.response = response

    @property
    def status_code(self):
        return self.response.status_code

    @property
    def headers(self):
        return self.response.headers

    @property
    async def text(self):
        return self.response.text

    @property
    async def content(self):
        return self.response.content

    @property
    async def json(self):
        return self.response.json()

    def iter_content(self, chunk_size=1024, decode_unicode=False):
        return self.response.iter_bytes()



import json

from linebot.__about__ import __version__
from linebot.exceptions import LineBotApiError
from linebot.http_client import HttpClient
from linebot.models import (
    Content,
    Error,
    MemberIds,
    MessageDeliveryBroadcastResponse,
    MessageDeliveryMulticastResponse,
    MessageDeliveryPushResponse,
    MessageDeliveryReplyResponse,
    MessageQuotaResponse,
    Profile,
    RichMenuResponse,
    MessageQuotaConsumptionResponse,
    IssueLinkTokenResponse,
    IssueChannelTokenResponse,
)

from .http_client import HttpXClient


class LineBotApiAsync(object):

    DEFAULT_API_ENDPOINT = 'https://api.line.me'
    DEFAULT_API_DATA_ENDPOINT = 'https://api-data.line.me'

    def __init__(
        self,
        channel_access_token,
        endpoint=DEFAULT_API_ENDPOINT,
        data_endpoint=DEFAULT_API_DATA_ENDPOINT,
        timeout=HttpClient.DEFAULT_TIMEOUT,
        http_client=None,
    ):
        self.endpoint = endpoint
        self.data_endpoint = data_endpoint
        self.headers = {
            'Authorization': 'Bearer ' + channel_access_token,
            'User-Agent': 'line-bot-sdk-python/' + __version__,
        }

        if http_client:
            self.http_client = http_client
        else:
            self.http_client = HttpXClient(timeout=timeout)

    async def close(self):
        await self.http_client.close()

    async def reply_message(self, reply_token, messages, timeout=None):
        if not isinstance(messages, (list, tuple)):
            messages = [messages]

        data = {
            'replyToken': reply_token,
            'messages': [message.as_json_dict() for message in messages],
        }

        await self._post(
            '/v2/bot/message/reply', data=json.dumps(data), timeout=timeout
        )

    async def push_message(self, to, messages, timeout=None):
        if not isinstance(messages, (list, tuple)):
            messages = [messages]

        data = {'to': to, 'messages': [message.as_json_dict() for message in messages]}

        await self._post('/v2/bot/message/push', data=json.dumps(data), timeout=timeout)

    async def get_rich_menu(self, rich_menu_id, timeout=None):
        response = await self._get(
            '/v2/bot/richmenu/{rich_menu_id}'.format(rich_menu_id=rich_menu_id),
            timeout=timeout,
        )

        json_resp = await response.json
        return RichMenuResponse.new_from_json_dict(json_resp)

    async def delete_rich_menu(self, rich_menu_id, timeout=None):
        await self._delete(
            '/v2/bot/richmenu/{rich_menu_id}'.format(rich_menu_id=rich_menu_id),
            timeout=timeout,
        )

    async def create_rich_menu(self, rich_menu, timeout=None):
        response = await self._post(
            '/v2/bot/richmenu', data=rich_menu.as_json_string(), timeout=timeout
        )

        json_resp = await response.json
        return json_resp.get('richMenuId')

    async def link_rich_menu_to_user(self, user_id, rich_menu_id, timeout=None):
        await self._post(
            '/v2/bot/user/{user_id}/richmenu/{rich_menu_id}'.format(
                user_id=user_id, rich_menu_id=rich_menu_id
            ),
            timeout=timeout,
        )

    async def link_rich_menu_to_users(self, user_ids, rich_menu_id, timeout=None):
        await self._post(
            '/v2/bot/richmenu/bulk/link',
            data=json.dumps({'userIds': user_ids, 'richMenuId': rich_menu_id}),
            timeout=timeout,
        )

    async def get_rich_menu_list(self, timeout=None):
        response = await self._get('/v2/bot/richmenu/list', timeout=timeout)

        lst_result = []
        json_resp = await response.json
        for richmenu in json_resp['richmenus']:
            lst_result.append(RichMenuResponse.new_from_json_dict(richmenu))

        return lst_result

    async def set_rich_menu_image(
        self, rich_menu_id, content_type, content, timeout=None
    ):
        await self._post(
            '/v2/bot/richmenu/{rich_menu_id}/content'.format(rich_menu_id=rich_menu_id),
            data=content,
            headers={'Content-Type': content_type},
            timeout=timeout,
            endpoint=self.data_endpoint
        )

    async def get_rich_menu_id_of_user(self, user_id, timeout=None):
        response = await self._get(
            '/v2/bot/user/{user_id}/richmenu'.format(user_id=user_id), timeout=timeout
        )

        json_resp = await response.json
        return json_resp.get('richMenuId')

    async def unlink_rich_menu_from_user(self, user_id, timeout=None):
        await self._delete(
            '/v2/bot/user/{user_id}/richmenu'.format(user_id=user_id), timeout=timeout
        )

    async def unlink_rich_menu_from_users(self, user_ids, timeout=None):
        await self._post(
            '/v2/bot/richmenu/bulk/unlink',
            data=json.dumps({'userIds': user_ids}),
            timeout=timeout,
        )

    async def get_rich_menu_image(self, rich_menu_id, timeout=None):
        response = await self._get(
            '/v2/bot/richmenu/{rich_menu_id}/content'.format(rich_menu_id=rich_menu_id),
            timeout=timeout,
            endpoint=self.data_endpoint
        )

        return Content(response)

    async def multicast(self, to, messages, timeout=None):
        if not isinstance(messages, (list, tuple)):
            messages = [messages]

        data = {'to': to, 'messages': [message.as_json_dict() for message in messages]}

        await self._post(
            '/v2/bot/message/multicast', data=json.dumps(data), timeout=timeout
        )

    async def get_profile(self, user_id, timeout=None):
        response = await self._get(
            '/v2/bot/profile/{user_id}'.format(user_id=user_id), timeout=timeout
        )

        json_resp = await response.json
        return Profile.new_from_json_dict(json_resp)

    async def get_group_member_profile(self, group_id, user_id, timeout=None):
        response = await self._get(
            '/v2/bot/group/{group_id}/member/{user_id}'.format(
                group_id=group_id, user_id=user_id
            ),
            timeout=timeout,
        )

        json_resp = await response.json
        return Profile.new_from_json_dict(json_resp)

    async def get_room_member_profile(self, room_id, user_id, timeout=None):
        response = await self._get(
            '/v2/bot/room/{room_id}/member/{user_id}'.format(
                room_id=room_id, user_id=user_id
            ),
            timeout=timeout,
        )

        json_resp = await response.json
        return Profile.new_from_json_dict(json_resp)

    async def get_group_member_ids(self, group_id, start=None, timeout=None):
        params = None if start is None else {'start': start}

        response = await self._get(
            '/v2/bot/group/{group_id}/members/ids'.format(group_id=group_id),
            params=params,
            timeout=timeout,
        )

        json_resp = await response.json
        return MemberIds.new_from_json_dict(json_resp)

    async def get_room_member_ids(self, room_id, start=None, timeout=None):
        params = None if start is None else {'start': start}

        response = await self._get(
            '/v2/bot/room/{room_id}/members/ids'.format(room_id=room_id),
            params=params,
            timeout=timeout,
        )

        json_resp = await response.json
        return MemberIds.new_from_json_dict(json_resp)

    async def get_message_content(self, message_id, timeout=None):
        response = await self._get(
            '/v2/bot/message/{message_id}/content'.format(message_id=message_id),
            endpoint=self.data_endpoint,
            stream=True,
            timeout=timeout,
        )

        return Content(response)

    async def leave_group(self, group_id, timeout=None):
        await self._post(
            '/v2/bot/group/{group_id}/leave'.format(group_id=group_id), timeout=timeout
        )

    async def leave_room(self, room_id, timeout=None):
        await self._post(
            '/v2/bot/room/{room_id}/leave'.format(room_id=room_id), timeout=timeout
        )

    async def broadcast(self, messages, notification_disabled=False, timeout=None):
        if not isinstance(messages, (list, tuple)):
            messages = [messages]

        data = {
            'messages': [message.as_json_dict() for message in messages],
            'notificationDisabled': notification_disabled,
        }

        await self._post(
            '/v2/bot/message/broadcast', data=json.dumps(data), timeout=timeout
        )

    async def cancel_default_rich_menu(self, timeout=None):
        await self._delete('/v2/bot/user/all/richmenu', timeout=timeout)

    async def set_default_rich_menu(self, rich_menu_id, timeout=None):
        await self._post(
            '/v2/bot/user/all/richmenu/{rich_menu_id}'.format(
                rich_menu_id=rich_menu_id
            ),

            timeout=timeout,
        )

    async def get_default_rich_menu(self, timeout=None):
        response = await self._get('/v2/bot/user/all/richmenu', timeout=timeout)

        json_resp = await response.json
        return json_resp.get('richMenuId')

    async def get_message_delivery_broadcast(self, date, timeout=None):
        response = await self._get(
            '/v2/bot/message/delivery/broadcast?date={date}'.format(date=date),
            timeout=timeout,
        )

        json_resp = await response.json
        return MessageDeliveryBroadcastResponse.new_from_json_dict(json_resp)

    async def get_message_delivery_reply(self, date, timeout=None):
        response = await self._get(
            '/v2/bot/message/delivery/reply?date={date}'.format(date=date),
            timeout=timeout,
        )

        json_resp = await response.json
        return MessageDeliveryReplyResponse.new_from_json_dict(json_resp)

    async def get_message_delivery_push(self, date, timeout=None):
        response = await self._get(
            '/v2/bot/message/delivery/push?date={date}'.format(date=date),
            timeout=timeout,
        )

        json_resp = await response.json
        return MessageDeliveryPushResponse.new_from_json_dict(json_resp)

    async def get_message_delivery_multicast(self, date, timeout=None):
        response = await self._get(
            '/v2/bot/message/delivery/multicast?date={date}'.format(date=date),
            timeout=timeout,
        )

        json_resp = await response.json
        return MessageDeliveryMulticastResponse.new_from_json_dict(json_resp)

    async def get_message_quota(self, timeout=None):
        response = await self._get('/v2/bot/message/quota', timeout=timeout)

        json_resp = await response.json
        return MessageQuotaResponse.new_from_json_dict(json_resp)

    async def get_message_quota_consumption(self, timeout=None):
        response = await self._get('/v2/bot/message/quota/consumption', timeout=timeout)

        json_resp = await response.json
        return MessageQuotaConsumptionResponse.new_from_json_dict(json_resp)

    async def issue_link_token(self, user_id, timeout=None):
        response = await self._post(
            '/v2/bot/user/{user_id}/linkToken'.format(user_id=user_id), timeout=timeout
        )

        json_resp = await response.json
        return IssueLinkTokenResponse.new_from_json_dict(json_resp)

    async def issue_channel_token(
        self, client_id, client_secret, grant_type='client_credentials', timeout=None
    ):
        response = await self._post(
            '/v2/oauth/accessToken',
            data={
                'client_id': client_id,
                'client_secret': client_secret,
                'grant_type': grant_type,
            },
            headers={'Content-Type': 'application/x-www-form-urlencoded'},
            timeout=timeout,
        )

        json_resp = await response.json
        return IssueChannelTokenResponse.new_from_json_dict(json_resp)

    async def revoke_channel_token(self, access_token, timeout=None):
        await self._post(
            '/v2/oauth/revoke',
            data={'access_token': access_token},
            headers={'Content-Type': 'application/x-www-form-urlencoded'},
            timeout=timeout,
        )

    async def _get(self, path, params=None, endpoint=None, headers=None, stream=False, timeout=None):
        if endpoint:
            url = endpoint + path
        else:
            url = self.endpoint + path

        if headers is None:
            headers = {}
        headers.update(self.headers)

        response = await self.http_client.get(
            url, headers=headers, params=params, stream=stream, timeout=timeout
        )

        await self.__check_error(response)
        return response

    async def _post(self, path, data=None, endpoint=None, headers=None, timeout=None):
        if endpoint:
            url = endpoint + path
        else:
            url = self.endpoint + path

        if headers is None:
            headers = {'Content-Type': 'application/json'}
        headers.update(self.headers)

        response = await self.http_client.post(
            url, headers=headers, data=data, timeout=timeout
        )
        await self.__check_error(response)
        return response

    async def _delete(self, path, data=None, endpoint=None, headers=None, timeout=None):
        if endpoint:
            url = endpoint + path
        else:
            url = self.endpoint + path

        if headers is None:
            headers = {}
        headers.update(self.headers)

        response = await self.http_client.delete(
            url, headers=headers, params=data, timeout=timeout
        )

        await self.__check_error(response)
        return response

    @staticmethod
    async def __check_error(response):
        if 200 <= response.status_code < 300:
            pass
        else:
            error = Error.new_from_json_dict((await response.json))
            raise LineBotApiError(response.status_code, error)



import asyncio
import inspect

from linebot.models.events import MessageEvent
from linebot.webhook import WebhookHandler
from linebot.utils import LOGGER


class WebhookHandlerAsync(WebhookHandler):
    def __init__(self, channel_secret):
        super().__init__(channel_secret)

    def add(self, event, message=None):
        def decorator(func):
            if not asyncio.iscoroutinefunction(func):
                raise TypeError("Handler must be a coroutine function")

            if isinstance(message, (list, tuple)):
                for it in message:
                    self.__add_handler(func, event, message=it)
            else:
                self.__add_handler(func, event, message=message)

            return func

        return decorator

    def default(self):
        def decorator(func):
            if not asyncio.iscoroutinefunction(func):
                raise TypeError("Handler must be a coroutine function")

            self._default = func
            return func

        return decorator

    async def handle(self, body, signature):
        payload = self.parser.parse(body, signature, as_payload=True)

        for event in payload.events:
            func = None
            key = None

            if isinstance(event, MessageEvent):
                key = self.__get_handler_key(event.__class__, event.message.__class__)
                func = self._handlers.get(key, None)

            if func is None:
                key = self.__get_handler_key(event.__class__)
                func = self._handlers.get(key, None)

            if func is None:
                func = self._default

            if func is None:
                LOGGER.info("No handler of " + key + " and no default handler")
            else:
                args_count = self.__get_args_count(func)
                if args_count == 0:
                    await func()
                elif args_count == 1:
                    await func(event)
                else:
                    await func(event, payload.destination)

    def __add_handler(self, func, event, message=None):
        key = self.__get_handler_key(event, message=message)
        self._handlers[key] = func

    @staticmethod
    def __get_args_count(func):
        arg_spec = inspect.getfullargspec(func)
        return len(arg_spec.args)

    @staticmethod
    def __get_handler_key(event, message=None):
        if message is None:
            return event.__name__
        else:
            return event.__name__ + "_" + message.__name__
